/*
 *  attacks.c - functions detecting attacks
 *
 *  Copyright (C) 1999 Hugo Haas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <time.h>
#include <stdlib.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "attacks.h"
#include "defines.h"
#include "log.h"

extern struct loginfo log;

/********************
 * Global functions *
 ********************/

/*
 * find_record
 *
 * Find the record corresponding to an IP address
 */

struct packet_record *find_record(struct packet_record *root, __u32 ipaddr) {
  struct packet_record *p;

  p = root;
  while (p != NULL) {
    if (p->ip == ipaddr)
      return(p);
    p = p-> next;
  }
  return(NULL);
}

/*
 * create_record
 *
 * Create a record
 */

struct packet_record *create_record(struct packet_record **root, time_t current, __u32 ipaddr) {
  struct packet_record *p;

  p = (struct packet_record *) malloc(sizeof(struct packet_record));
  if (p == NULL)
    fatal();
  p->ip = ipaddr;
  p->latest = p->started = current;
  p->desc.count = 1;
  p->triggered = FALSE;
  p->prev = NULL;
  p->next = *root;
  return(*root = p);
}

/*
 * remove_record
 *
 * Remove a record from a structure
 */

void remove_record(struct packet_record **root, struct packet_record *r) {

  if (r == *root)
    *root = r->next;
  if (r->prev)
    r->prev->next = r->next;
  if (r->next)
    r->next->prev = r->prev;
  free(r);
}

/*
 * destroy_structure
 *
 * Free a set of records
 */

void destroy_structure(struct packet_record **root) {
  struct packet_record *p;

  while (*root != NULL) {
    p = (*root)->next;
    free(*root);
    *root = p;
  }
}

/*
 * timeout
 *
 * Remove old record
 */

void timeout(struct packet_record **root, int timeout) {
  struct packet_record *p,*q;
  time_t current;

  time(&current);
  p = *root;

  while (p != NULL) {
    q = p->next;
    if ((current-p->latest) > timeout)
      remove_record(root, p);
    p = q;
  }  
}

/************************
 * Ping flood detection *
 ***********************/

struct packet_record *ping_rec;
int ping_threshold = 5;
int ping_ignore = 5;
int ping_flush = 60;
time_t ping_last = 0;

/*
 * record_ping
 *
 * Record an ICMP echo record request
 */

int record_ping(__u32 ipaddr) {
  struct packet_record *record;
  time_t current;

  time(&current);

  if ((current-ping_last) > ping_flush)
    timeout(&ping_rec, ping_ignore);

  record = find_record(ping_rec, ipaddr);
  if (record == NULL) {
    create_record(&ping_rec, current, ipaddr);
    return(TRUE);
  }
  else {
    /* Ping flood detected? */
    if (record->triggered == TRUE) {
      if (current-record->latest >= ping_ignore) {
	remove_record(&ping_rec, record);
	return(TRUE);
      }
      record->latest = current;
      return(FALSE);
    }
    if ((current-record->latest) < 1) {
      record->desc.count++;
      /* Triggered? */
      if (record->desc.count >= ping_threshold) {
	struct in_addr i;
	i.s_addr = ipaddr;
	record->latest = current;
	record->triggered = TRUE;
	log.log(log.level_or_fd, "Attack! Ping flood from %s detected", inet_ntoa(i));
	return(FALSE);
      }
    }
    else {
      record->desc.count = 1;
    }
  }
  record->latest = current;
  return(TRUE);
}

/*******************************
 * Initialisation and clean-up *
 ******************************/

/*
 * init_attack_records
 *
 * Initialize the structures
 */

void init_attack_records(void) {

  destroy_structure(&ping_rec);
}
